@@ -1,4 +1,7 @@
-use std::io::{self, Read};
+use std::{
+    convert::TryFrom,
+    io::{self, Read},
+};
 
 use noodles_sam::record::data::field::Tag;
 
@@ -6,17 +9,9 @@ pub fn read_tag<R>(reader: &mut R) -> io::Result<Tag>
 where
     R: Read,
 {
-    use std::str;
-
     let mut buf = [0; 2];
     reader.read_exact(&mut buf)?;
-
-    str::from_utf8(&buf)
-        .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-        .and_then(|s| {
-            s.parse()
-                .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-        })
+    Tag::try_from(buf).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
 }
 
 #[cfg(test)]
